
import numpy as np

from scipy.sparse.linalg import eigsh
from scipy import sparse

from tqdm.notebook import tqdm

from .mass_tet_basis import mass_tet10


def construct_loce(org, conec, n_dof_per_node=3):
    loce = []
    insert = np.tile(np.arange(n_dof_per_node), conec.shape[1])
    for element in conec:
        loce.append(np.repeat(element, n_dof_per_node)*n_dof_per_node + insert)
    return np.array(loce)

def _assemble_B_matrix_tet10(N_diff):
    B = np.zeros((N_diff.shape[0], 6, 30))
    for j in range(10):
        B[:, 0, j*3+0] = N_diff[:, 0, j]
        B[:, 1, j*3+1] = N_diff[:, 1, j]
        B[:, 2, j*3+2] = N_diff[:, 2, j]

        B[:, 3, j*3+0] = N_diff[:, 1, j]
        B[:, 3, j*3+1] = N_diff[:, 0, j]
        B[:, 4, j*3+1] = N_diff[:, 2, j]
        B[:, 4, j*3+2] = N_diff[:, 1, j]
        B[:, 5, j*3+0] = N_diff[:, 2, j]
        B[:, 5, j*3+2] = N_diff[:, 0, j]
    return B

def H_matrix(E, nu):
    d1 = E*(1-nu)/((1+nu)*(1-2*nu))
    d2 = nu*E/((1+nu)*(1-2*nu))
    d3 = E/(2*(1+nu))

    H = np.zeros((E.shape[0], 6, 6))
    H[:, 0, 0] = d1
    H[:, 1, 1] = d1
    H[:, 2, 2] = d1
    H[:, 3, 3] = d3
    H[:, 4, 4] = d3
    H[:, 5, 5] = d3
    H[:, 0, 1] = d2
    H[:, 0, 2] = d2
    H[:, 1, 0] = d2
    H[:, 1, 2] = d2
    H[:, 2, 0] = d2
    H[:, 2, 1] = d2
    return H

def Jacobi_det(J):
    J_det = J[:, 0, 0]*((J[:, 1, 1] * J[:, 2, 2]) - (J[:, 1, 2] * J[:, 2, 1]))\
                - J[:, 0, 1]*((J[:, 1, 0] * J[:, 2, 2]) - (J[:, 1, 2] * J[:, 2, 0]))\
                + J[:, 0, 2]*((J[:, 1, 0] * J[:, 2, 1]) -
                              (J[:, 1, 1] * J[:, 2, 0]))
    return J_det

def Jacobi_inv(J, J_det):
    J_t = np.transpose(J, (0, 2, 1))
    J_inv = np.zeros_like(J)
    J_inv[:, 0, 0] = ((J_t[:, 1, 1] * J_t[:, 2, 2]) -
                      (J_t[:, 1, 2] * J_t[:, 2, 1])) / J_det
    J_inv[:, 0, 1] = -((J_t[:, 1, 0] * J_t[:, 2, 2]) -
                       (J_t[:, 1, 2] * J_t[:, 2, 0])) / J_det
    J_inv[:, 0, 2] = ((J_t[:, 1, 0] * J_t[:, 2, 1]) -
                      (J_t[:, 1, 1] * J_t[:, 2, 0])) / J_det
    J_inv[:, 1, 0] = -((J_t[:, 0, 1] * J_t[:, 2, 2]) -
                       (J_t[:, 0, 2] * J_t[:, 2, 1])) / J_det
    J_inv[:, 1, 1] = ((J_t[:, 0, 0] * J_t[:, 2, 2]) -
                      (J_t[:, 0, 2] * J_t[:, 2, 0])) / J_det
    J_inv[:, 1, 2] = -((J_t[:, 0, 0] * J_t[:, 2, 1]) -
                       (J_t[:, 0, 1] * J_t[:, 2, 0])) / J_det
    J_inv[:, 2, 0] = ((J_t[:, 0, 1] * J_t[:, 1, 2]) -
                      (J_t[:, 0, 2] * J_t[:, 1, 1])) / J_det
    J_inv[:, 2, 1] = -((J_t[:, 0, 0] * J_t[:, 1, 2]) -
                       (J_t[:, 0, 2] * J_t[:, 1, 0])) / J_det
    J_inv[:, 2, 2] = ((J_t[:, 0, 0] * J_t[:, 1, 1]) -
                      (J_t[:, 0, 1] * J_t[:, 1, 0])) / J_det
    return J_inv

class Tetrahedron:
    def __init__(self, org, conec, Young, Density, Poisson, calc_n_freq=20, dof_mask=None, added_masses=None, mass_locations=None, org_rotation=None, lumped=False):
        """Initialize the tetrahedron model.

        Parameters
        ----------
        org : array_like
            Nodes.
        conec : array_like
            Elements.
        Young : array_like
            Stiffnesses of the elements.
        Density : array_like
            Density of the elements.
        Poisson : array_like
            Poisson ratio of the material.
        calc_n_freq : int
            How many natural frequencies to compute.
        dof_mask : array_like
            Mask the degrees of freedom that are rigidly fixated.
        added_masses : array_like
            Masses of point masses.
        mass_locations : array_like
            Locations of the point masses.
        org_rotation : array_like
            Rotation around the x, y, and z axes (in radians). This is a temporary rotation
            for rigid fixation that is not in-line with one of the axes.
        """
        self.n_dof_node = 3
        self.calc_n_freq = calc_n_freq
        
        self.ro = Density
        self.E = Young
        self.nu = Poisson
        self.org = org
        self.conec = conec
        self.loce = construct_loce(org, conec, self.n_dof_node)
        self.dof_mask = dof_mask

        self.added_masses = added_masses
        self.mass_locations = mass_locations
        self.org_rotation = org_rotation

        self.m_base = mass_tet10()
        
        self.n_nodes = self.org.shape[0]  # število vozlišč v mreži
        self.n_el = self.conec.shape[0]  # število elementov v mreži
        self.n_dof = self.n_dof_node * self.n_nodes  # število prostostnih stopenj

        self.n_dof_element = self.conec.shape[1] * self.n_dof_node
        self.Jg = np.repeat(self.loce.flatten(), self.n_dof_element)
        self.Ig = np.tile(self.loce, self.n_dof_element).flatten()
        
        if type(self.E) in [float, int, np.float64]:
            self.E = np.repeat(self.E, self.n_el)
        if type(self.ro) in [float, int, np.float64]:
            self.ro = np.repeat(self.ro, self.n_el)

        if lumped:
            self.K, self.M = self.assemble_matrices_lumped(self.E, self.ro)
        else:
            self.K, self.M = self.assemble_matrices(self.E, self.ro)

        if self.org_rotation is not None:
            pass

        if self.added_masses is not None and self.mass_locations is not None:
            self.add_point_masses()
    
    # def add_point_masses(self):
    #     for i, m in zip(self.mass_locations, self.added_masses):
    #         # self.M[np.ix_(self.loce[i], self.loce[i])][0, 0] += m
    #         for _ in range(self.loce.shape[1]):
    #             self.M[self.loce[i][_], self.loce[i][_]] += (m/self.conec.shape[1])
    #         # self.M[self.loce[i][0], self.loce[i][0]] += m
    
    def add_point_masses(self):
        for i, m in zip(self.mass_locations, self.added_masses):
            add_loc = i*3 + np.array([0, 1, 2])
            self.M[add_loc, add_loc] += m
                # self.M[np.ix_(add_loc, add_loc)] += np.zeros((3, 3)) * m


    def assemble_matrices_lumped(self, E, ro):
        """
        Assembling the element matrices into the global mass and stiffness matrix.
        
        To calculate the derivative of the matrix with respect to rho or E, 
        simply set that parameter to 1. Everything else is calculated the same way.
        
        Parameters
        ----------
        rho : float or array_like
            Density of the elements.
        E : float or array_like
            Young's modulus (stiffness) of the elements.
        n_el : int
            Number of elements.
        """
        Kg = np.zeros((self.n_el * self.n_dof_element**2))
        Mg = np.zeros_like(Kg)

        # Assemble H matrix
        H = H_matrix(E, self.nu)

        # Gauss quadrature locations and weights
        loc1 = 0.58541020
        loc2 = 0.13819660
        xis = [loc1, loc2, loc2, loc2]*2
        etas = [loc2, loc1, loc2, loc2]*2
        zetas = [loc2, loc2, loc1, loc2]*2
        w = 0.041666667

        # Iterate over integration points
        for _, (xi, eta, zeta) in enumerate(zip(xis, etas, zetas)):
            lam = 1 - xi - eta - zeta
            N_diffNat = np.array([[-1+4*xi, 0, 1-4*lam, 0, 4*eta, -4*eta, 4*(lam-xi), 4*zeta, 0, -4*zeta],
                                  [0, -1+4*eta, 1-4*lam, 0, 4*xi, 4 *
                                      (lam-eta), -4*xi, 0, 4*zeta, -4*zeta],
                                  [0, 0, 1-4*lam, -1+4*zeta, 0, -4*eta, -4*xi, 4*xi, 4*eta, 4*(lam-zeta)]])

            J = N_diffNat @ self.org[self.conec]

            # Jacobi determinant
            J_det = Jacobi_det(J)

            if _ >= 4:
                # Mass matrix
                Mg += np.transpose(ro[:, None, None] * J_det[:, None, None]/60/4 * np.identity(30)[None, :, :], (0, 2, 1)).flatten()
            else:
                # Stiffness matrix
                # Jacobi inverse
                J_inv = Jacobi_inv(J, J_det)
                
                # Shape functions
                N_diff = J_inv @ N_diffNat

                # B matrix
                B = _assemble_B_matrix_tet10(N_diff)

                # Add to stiffness matrix
                Kg += np.transpose(w * J_det[:, None, None] * (
                    np.transpose(B, (0, 2, 1)) @ H) @ B, (0, 2, 1)).flatten()
        
        M = sparse.csc_matrix(
            (Mg, (self.Ig, self.Jg)), shape=(self.n_dof, self.n_dof))
        K = sparse.csc_matrix(
            (Kg, (self.Ig, self.Jg)), shape=(self.n_dof, self.n_dof))

        del Kg
        del Mg

        if self.dof_mask is not None:
            K = K[self.dof_mask][:, self.dof_mask]
            M = M[self.dof_mask][:, self.dof_mask]
        
        return K, M


    def assemble_matrices(self, E, ro):
        """
        Assembling the element matrices into the global mass and stiffness matrix.
        
        To calculate the derivative of the matrix with respect to rho or E, simply set that parameter to 1. Everything else is calculated the same way.
        
        Parameters
        ----------
        rho : float or array_like
            Density of the elements.
        E : float or array_like
            Young's modulus (stiffness) of the elements.
        n_el : int
            Number of elements.
        """
        Kg = np.zeros((self.n_el * self.n_dof_element**2))
        Mg = np.zeros_like(Kg)

        # Assemble H matrix
        H = H_matrix(E, self.nu)

        # Gauss quadrature locations and weights
        loc1 = 0.58541020
        loc2 = 0.13819660
        xis = [loc1, loc2, loc2, loc2, 1/4]
        etas = [loc2, loc1, loc2, loc2, 1/4]
        zetas = [loc2, loc2, loc1, loc2, 1/4]
        w = 0.041666667

        # Iterate over integration points
        for _, (xi, eta, zeta) in enumerate(zip(xis, etas, zetas)):
            lam = 1 - xi - eta - zeta
            N_diffNat = np.array([[-1+4*xi, 0, 1-4*lam, 0, 4*eta, -4*eta, 4*(lam-xi), 4*zeta, 0, -4*zeta],
                                  [0, -1+4*eta, 1-4*lam, 0, 4*xi, 4 *
                                      (lam-eta), -4*xi, 0, 4*zeta, -4*zeta],
                                  [0, 0, 1-4*lam, -1+4*zeta, 0, -4*eta, -4*xi, 4*xi, 4*eta, 4*(lam-zeta)]])

            J = N_diffNat @ self.org[self.conec]

            # Jacobi determinant
            J_det = Jacobi_det(J)

            if _ == 4:
                # Mass matrix
                Mg = np.transpose(ro[:, None, None] * J_det[:, None, None]/2520 * self.m_base[None, :, :], (0, 2, 1)).flatten()
            else:
                # Stiffness matrix
                # Jacobi inverse
                J_inv = Jacobi_inv(J, J_det)
                
                # Shape functions
                N_diff = J_inv @ N_diffNat

                # B matrix
                B = _assemble_B_matrix_tet10(N_diff)

                # Add to stiffness matrix
                Kg += np.transpose(w * J_det[:, None, None] * (
                    np.transpose(B, (0, 2, 1)) @ H) @ B, (0, 2, 1)).flatten()
        
        M = sparse.csc_matrix(
            (Mg, (self.Ig, self.Jg)), shape=(self.n_dof, self.n_dof))
        K = sparse.csc_matrix(
            (Kg, (self.Ig, self.Jg)), shape=(self.n_dof, self.n_dof))

        del Kg
        del Mg

        if self.dof_mask is not None:
            K = K[self.dof_mask][:, self.dof_mask]
            M = M[self.dof_mask][:, self.dof_mask]
        
        return K, M
    
    def solve(self):
        """
        Solve the eigenvalue problem.
        
        Parameters
        ----------
        dof_mask : array_like, optional
            Degrees of freedom that are not fixed. If None, the body is free-free.
        """
        self.eigval, self.eigvec = eigsh(self.K, M=self.M, k=self.calc_n_freq, sigma=0, which='LM')

        eig_omega = np.sqrt(self.eigval)
        eig_omega[eig_omega != eig_omega] = 0  # nan element are 0
        self.nat_freq = eig_omega / (2*np.pi)
        # self.A = np.array([v.reshape(self.org.shape) for i, v in enumerate(self.eigvec.T)]) # reshape the eigenvectors

        if self.dof_mask is not None:
            eigvec_full = np.zeros((self.org.flatten().shape[0], self.eigvec.shape[1]))
            eigvec_full[self.dof_mask] = self.eigvec
            self.eigvec_full = eigvec_full
        else:
            self.eigvec_full = self.eigvec

        try:
            self.A = self.reshape_eigenvectors(self.eigvec_full, self.org.shape)
        except:
            pass

    def reshape_eigenvectors(self, eigvec, shape):
        return np.array([v.reshape(shape) for i, v in enumerate(eigvec.T)])
        
    def matrix_derivative(self, derivative_E_ind=None, derivative_ro_ind=None, diff_step=1e-5):
        if derivative_E_ind is not None:
            E = self.E.copy()
            step = np.mean(E[derivative_E_ind] * diff_step)
            
            E[derivative_E_ind] = E[derivative_E_ind] + step
            
            K_, M_ = self.assemble_matrices(E, self.ro)
            
            self.K_diff = (K_ - self.K) / step
            self.M_diff = sparse.csc_matrix(self.K_diff.shape)
            
        elif derivative_ro_ind is not None:
            ro = self.ro.copy()
            step = np.mean(ro[derivative_ro_ind] * diff_step)
            
            ro[derivative_ro_ind] = ro[derivative_ro_ind] + step
            
            K_, M_ = self.assemble_matrices(self.E, ro)
            
            self.M_diff = (M_ - self.M) / step
            self.K_diff = sparse.csc_matrix(self.M_diff.shape)
            
    def split_A(self, j, k_pivot=None):
        """Compute dynamic matrix A and performe LU split.
        
        Matrix is used for v_j computation page 25, Eq. (2.65) in [1].
        
        [1] Friswell, Mottershead. Finite Element Model Updating in Structural
            Dynamics.
        """
        if k_pivot is None:
            k_pivot = np.argmax(np.abs(self.eigvec[:, j]))
            
        A = (self.K - self.eigval[j]*self.M)
#         A = A.toarray()
#         A = np.insert(A, k_pivot, 0, axis=0)
#         A = np.insert(A, k_pivot, 0, axis=1)
        
        A._shape = (A.shape[0]+1, A.shape[1]+1)
        A.indptr = np.insert(A.indptr, k_pivot, A.indptr[k_pivot])
        A.indices[A.indices >= k_pivot] += 1
        A[k_pivot, k_pivot] = 1
        
        lu_A = sparse.linalg.splu(A)
        return lu_A
    
    def S_eigval(self, j):
        return self.eigvec[:, j] @ (self.K_diff - self.eigval[j]*self.M_diff) @ self.eigvec[:, j]
    
    def S_matrix(self, eig_ind, derivative_E_ind, compute_vector_sensitivity=True, coor_inds=None, diff_step=1e-5):
        dMs = []
        dKs = []

        if coor_inds is not None:
            coor_inds1 = construct_loce(None, np.array([coor_inds]), 3)

        S_mtx_val = np.zeros((len(eig_ind), len(derivative_E_ind)))
        S_mtx_vec = np.zeros((self.eigvec.shape[0], len(eig_ind), len(derivative_E_ind)))

        for i1, j in enumerate(tqdm(eig_ind, leave=False)):
            if compute_vector_sensitivity:
                k_pivot = np.argmax(np.abs(self.eigvec[:, j]))
                lu_A = self.split_A(j, k_pivot)

            for i2, dEi in enumerate(tqdm(derivative_E_ind, leave=False)):
                # Assemble derivative matrices
                if i1 == 0:
        #             self.assemble_matrices(derivative_E_ind=dEi)
                    self.matrix_derivative(derivative_E_ind=dEi, diff_step=diff_step)
                    dMs.append(self.M_diff)
                    dKs.append(self.K_diff)
                else:
                    self.M_diff = dMs[i2]
                    self.K_diff = dKs[i2]

                # Eigenvalue sensitivity
                S_eigval = self.S_eigval(j=j)

                if compute_vector_sensitivity:
                    # Eigenvector sensitivity
                    f_j = -(self.K_diff - self.eigval[j]*self.M_diff - S_eigval * self.M) @ self.eigvec[:, j]
                    f_j = np.insert(f_j, k_pivot, 0)

                    v_j = lu_A.solve(f_j)
                    v_j = np.delete(v_j, k_pivot)

                    c_j = -self.eigvec[:, j] @ self.M @ v_j - 1/2*self.eigvec[:, j] @ self.M_diff @ self.eigvec[:, j]

                    S_eigvec = v_j + c_j*self.eigvec[:, j]

                    S_mtx_vec[:, i1, i2] = S_eigvec

                S_mtx_val[i1, i2] = S_eigval

        if compute_vector_sensitivity:
            if coor_inds is not None:
                S_mtx_vec = S_mtx_vec[coor_inds1]

            S_mtx_vec = S_mtx_vec.reshape(-1, len(derivative_E_ind))
            S_mtx = np.concatenate((S_mtx_val, S_mtx_vec))

        else:
            S_mtx = S_mtx_val
            
        return S_mtx    

    # Working version, for reference.
    # def S_matrix(self, eig_ind, derivative_E_ind, compute_vector_sensitivity=True, coor_mask=None):
    #     dMs = []
    #     dKs = []

    #     S_mtx_val = np.zeros((len(eig_ind), len(derivative_E_ind)))
    #     S_mtx_vec = np.zeros((len(eig_ind)*self.eigvec.shape[0], len(derivative_E_ind)))

    #     for i1, j in enumerate(tqdm(eig_ind)):
    #         if compute_vector_sensitivity:
    #             k_pivot = np.argmax(np.abs(self.eigvec[:, j]))
    #             lu_A = self.split_A(j, k_pivot)

    #         for i2, dEi in enumerate(tqdm(derivative_E_ind, leave=False)):
    #             # Assemble derivative matrices
    #             if i1 == 0:
    #     #             self.assemble_matrices(derivative_E_ind=dEi)
    #                 self.matrix_derivative(derivative_E_ind=dEi)
    #                 dMs.append(self.M_diff)
    #                 dKs.append(self.K_diff)
    #             else:
    #                 self.M_diff = dMs[i2]
    #                 self.K_diff = dKs[i2]

    #             # Eigenvalue sensitivity
    #             S_eigval = self.S_eigval(j=j)

    #             if compute_vector_sensitivity:
    #                 # Eigenvector sensitivity
    #                 f_j = -(self.K_diff - self.eigval[j]*self.M_diff - S_eigval * self.M) @ self.eigvec[:, j]
    #                 f_j = np.insert(f_j, k_pivot, 0)

    #                 v_j = lu_A.solve(f_j)
    #                 v_j = np.delete(v_j, k_pivot)

    #                 c_j = -self.eigvec[:, j] @ self.M @ v_j - 1/2*self.eigvec[:, j] @ self.M_diff @ self.eigvec[:, j]

    #                 S_eigvec = v_j + c_j*self.eigvec[:, j]


    #                 S_mtx_vec[i1*self.eigvec.shape[0]:(i1+1)*self.eigvec.shape[0], i2] = S_eigvec

    #             S_mtx_val[i1, i2] = S_eigval

    #     S_mtx = np.concatenate((S_mtx_val, S_mtx_vec))
    #     return S_mtx


import os
import sys
import numpy as np
import matplotlib.pyplot as plt

current_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(current_path)

from sdypy.model import Beam


n_elements = 20
beam_length = 500 # mm
beam_ro = 7850 * 1e-12
beam_Young = 180 * 1e3

profile_height = 15
profile_width = 30

length = np.ones(n_elements) * beam_length/(n_elements)
density = np.ones(n_elements) * beam_ro
Young = np.ones(n_elements) * beam_Young
n_nodes = n_elements + 1

beam_obj = Beam(org=None, conec=None, 
        length=length, width=profile_width, 
        height=profile_height, density=density, 
        Young=Young, n_nodes=n_nodes, added_masses=None, mass_locations=None)

nat_freq, vec = beam_obj.solve()

print(nat_freq)

fig, axs = plt.subplots(3, 2, figsize=(10, 7))
for i in range(6):
    axs.flatten()[i].plot(vec[::2, i])
    axs.flatten()[i].set_title(f'Mode {i+1} ({nat_freq[i]:.2f} Hz)')
plt.tight_layout()
plt.show()


import os
import numpy as np
import pyvista as pv
from pyvistaqt import BackgroundPlotter

def generate_quadrilateral_mesh(path_to_step, element_size, path_to_mesh=None, show_mesh="None"):
    """Generate quadrilateral mesh (semi-structured).
    
    Parameters
    ----------
    path_to_step : str
        Path to the STEP file.
    element_size : float
        Desired element size in STEP units.
    path_to_mesh : str
        Save path for the mesh. If not given, the mesh will not be saved.
    show_mesh : bool
        If True, display the mesh.
    
    Returns
    -------
    nodes : np.ndarray
        Node coordinates.
    elements : np.ndarray
        Element connectivity.
    """
    import gmsh
    model_name = os.path.basename(path_to_step).split(".")[0]

    # Initialize GMSH
    gmsh.initialize()

    # Add a new model named 'imported_model'
    gmsh.model.add(model_name)

    # Import the STEP file
    gmsh.model.occ.importShapes(path_to_step)

    # Synchronize to ensure all imported geometries are available in the model
    gmsh.model.occ.synchronize()

    # Identify surfaces to be meshed
    # You can list all surfaces and then select the desired ones
    entities = gmsh.model.getEntities(dim=2)

    gmsh.model.mesh.field.add("MathEval", 1)
    gmsh.model.mesh.field.setString(1, "F", f"{element_size}")
    gmsh.model.mesh.field.setAsBackgroundMesh(1)

    # Set mesh options for 2D quadrilateral semi-structured mesh
    gmsh.option.setNumber("Mesh.Algorithm", 8)  # Use the transfinite algorithm for semi-structured meshes
    gmsh.option.setNumber("Mesh.RecombineAll", 1)  # Recombine into quadrilaterals

    # Optionally set transfinite meshing parameters (if you have specific preferences)
    # For each surface, define the transfinite algorithm parameters
    for entity in entities:
        gmsh.model.mesh.setTransfiniteSurface(entity[1])
        gmsh.model.mesh.setRecombine(2, entity[1])

    # Generate the 2D mesh
    gmsh.model.mesh.generate(2)

    # Save the mesh to a file
    if path_to_mesh:
        gmsh.write(path_to_mesh)

    node_tags, node_coords, _ = gmsh.model.mesh.getNodes()
    element_types, element_tags, element_node_tags = gmsh.model.mesh.getElements()
    nodes = np.reshape(node_coords, (-1, 3))
    i = np.argmin(np.abs(element_types - 3)) # element type 3 represent quadrilaterals
    elements = np.reshape(element_node_tags[i], (-1, 4)) - 1

    if show_mesh == "gmsh":
        gmsh.fltk.run()
    elif show_mesh == "pyvista":
        mesh = pv.PolyData(nodes)
        # Add cells (elements) to the mesh
        faces = []
        for element in elements:
            face = np.hstack([[4], element])  # Add number of points per cell (4 for quadrilateral)
            faces.append(face)
        faces = np.hstack(faces).astype(np.int64)
        mesh.faces = faces

        # Visualize the mesh with displacement coloring
        plotter = BackgroundPlotter()
        plotter.add_mesh(mesh, show_edges=True, edge_color='black')
        plotter.show_grid()
        plotter.show()
    else:
        pass

    # Finalize GMSH
    gmsh.finalize()

    return nodes, elements


import os
import sys
import tomli
import tomli_w
import argparse

package_name = "sdypy/model"

def synchronize_version():
    print("Synchronizing version (pyproject.toml and __init__.py)...")

    # Read the version from pyproject.toml
    with open("pyproject.toml", "rb") as f:
        pyproject = tomli.load(f)
    
    version_toml = pyproject["project"]["version"]

    # Read the __init__.py
    with open(f"{package_name}/__init__.py", "r") as f:
        init = f.readlines()

    # Replace the version with the one from pyproject.toml
    for i, line in enumerate(init):
        if "__version__" in line:
            init[i] = "__version__ = " + f'"{version_toml}"' + "\n"
    init = "".join(init)

    # Write the new __init__.py
    with open(f"{package_name}/__init__.py", "w") as f:
        f.write(init)

    # Update docs/source/conf.py
    with open("docs/source/conf.py", "r", encoding="utf8") as f:
        conf = f.readlines()

    for i, line in enumerate(conf):
        if "version = " in line and not line.strip().startswith("#"):
            conf[i] = f"version = '{version_toml.rsplit('.', 1)[0]}'\n"
        elif "release = " in line and not line.strip().startswith("#"):
            conf[i] = f"release = '{version_toml}'\n"

    # Write the new conf.py
    with open("docs/source/conf.py", "w", encoding="utf8") as f:
        f.write("".join(conf))

def set_version(version):
    with open("pyproject.toml", "rb") as f:
        pyproject = tomli.load(f)
    pyproject["project"]["version"] = version
    with open("pyproject.toml", "wb") as f:
        tomli_w.dump(pyproject, f)

def bump_version(bump):
    with open("pyproject.toml", "rb") as f:
        pyproject = tomli.load(f)
    version = pyproject["project"]["version"]
    version_parts = version.split(".")
    if bump == "patch":
        version_parts[2] = str(int(version_parts[2]) + 1)
    elif bump == "minor":
        version_parts[1] = str(int(version_parts[1]) + 1)
        version_parts[2] = "0"
    elif bump == "major":
        version_parts[0] = str(int(version_parts[0]) + 1)
        version_parts[1] = "0"
        version_parts[2] = "0"
    else:
        raise ValueError(f"Invalid bump type: {bump}")
    
    version = ".".join(version_parts)
    set_version(version)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--bump", default="", choices=["patch", "minor", "major"], help="Bump the version of the package.")
    parser.add_argument("--set-version", type=str, help="Set the version of the package.")
    args = parser.parse_args()

    if args.set_version:
        set_version(args.set_version)

    elif args.bump:
        bump_version(args.bump)

    synchronize_version()



import os
import sys
import numpy as np
import meshio
from pyvistaqt import BackgroundPlotter
import pyvista as pv
from scipy.sparse.linalg import eigsh
# os.environ['QT_QPA_PLATFORM'] = 'xcb' # if you run wayland and pyqt6, uncomment this

current_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(current_path)

from sdypy.model import Tetrahedron

def pad_facets(facets):
    return np.hstack((np.ones((facets.shape[0], 1))*facets.shape[1], facets)).astype(int)

# Load mesh
try:
    path_mesh = 'data/Beam - simple.msh'
    mesh = meshio.read(path_mesh)
except:
    path_mesh = 'examples/data/Beam - simple.msh'
    mesh = meshio.read(path_mesh)

org = mesh.points
org_orig = np.copy(org)
conec = mesh.cells_dict['tetra10']

# Create triangles
inds = np.array([[2, 1, 0, 4, 5, 6],
                 [1, 3, 0, 6, 7, 8],
                 [3, 2, 0, 5, 7, 9],
                 [2, 3, 1, 4, 8, 9]])

triangles = []
for element in conec:
    for i in inds:
        triangles.append(element[i])
triangles = np.unique(triangles, axis=0)
triangles_1 = pad_facets(triangles[:, :3])

# Plot the mesh
plotter = BackgroundPlotter()
grid = pv.PolyData(org, triangles_1)
grid_points = pv.PolyData(org)

plotter.add_mesh(grid_points, render_points_as_spheres=True, opacity=1, color='blue', point_size=7)
plotter.add_mesh(grid, render_points_as_spheres=True, opacity=1, color='white', show_edges=True)
plotter.add_axes()


# Assemble matrices
parameters = {
    'Young': np.ones((conec.shape[0])) * 190000,
    'Density': np.ones((conec.shape[0])) * 7850e-12,
    'Poisson': 0.3
}

tet_obj = Tetrahedron(org, conec, parameters['Young'], parameters['Density'], parameters['Poisson'], 60)

# Solve the eigenvalue problem
eigenvalues, eigenvectors = eigsh(tet_obj.K, M=tet_obj.M, k=20, sigma=0, which='LM')

nat_freq = np.sqrt(eigenvalues) / (2 * np.pi)

# %%
# Plot the modes
n_plots = 6
rows = int(np.floor(np.sqrt(n_plots)))
cols = int(n_plots // rows)
shape = (rows, cols)

# Initialize a plotter with 6 subplots (2 rows, 3 columns)
plotter = BackgroundPlotter(shape=shape)

mode_start = 6
for ii in range(int(rows*cols)):
    mode_number = mode_start + ii
    modal_shape = eigenvectors[:, mode_number].reshape(-1, 3)


    # Scale the modal shape for better visualization
    # scale_factor = 1e-2  # Adjust this as needed
    scale_factor = np.linalg.norm(org)/np.linalg.norm(modal_shape)/10
    displaced_nodes = org + scale_factor * modal_shape

    # Compute the norm of the displacements for coloring
    displacement_norm = np.linalg.norm(modal_shape, axis=1)

    # Select the subplot to work on
    plotter.subplot(ii // cols, ii % cols)

    # Create a mesh for the current mode
    mesh = pv.PolyData(displaced_nodes)

    triangles = []
    for element in conec:
        for i in inds:
            triangles.append(element[i])
    mesh.faces = pad_facets(np.array(triangles)[:, :3])

    # Add displacement norm as point data for coloring
    displacement_norm = np.linalg.norm(modal_shape, axis=1)  # Assuming displacements is a 3D array
    mesh["Displacement"] = displacement_norm

    # Add the mesh to the plotter
    plotter.add_mesh(mesh, show_edges=True, scalars="Displacement", cmap="viridis", edge_color='black')
    plotter.add_text(f"Mode {mode_number} ({np.sqrt(eigenvalues[mode_number])/(2*np.pi):.2f} Hz)", font_size=12)
    plotter.show_grid()

# Show the plotter with all subplots
plotter.show()
input("Press any key to exit.")
# %%



__version__ = "0.1.4"

import pyLump as lumped
from .shell.shell import Shell
from .tetrahedron.tet10 import Tetrahedron
from .beam.beam import Beam
from .eigenvalue_solution import solve_eigenvalue

from . import mesh


from scipy.sparse.linalg import eigsh
from scipy import sparse
from scipy.linalg import eigh
import numpy as np


def solve_eigenvalue(K, M, n_modes=None, convert_to_hz=False):
    """
    Solve the eigenvalue problem.

    Parameters
    ----------
    K : ndarray or sparse
        Stiffness matrix.
    M : ndarray or sparse
        Mass matrix.
    n_modes : int or None
        Number of modes to compute. If None, all modes are computed.

    Returns
    -------
    eigenvalues : ndarray
        Eigenvalues.
    eigenvectors : ndarray
        Eigenvectors.
    """
    if sparse.issparse(K) and sparse.issparse(M):
        if n_modes is None:
            n_modes = K.shape[0] - 1

        eigenvalues, eigenvectors = eigsh(K, M=M, k=n_modes, sigma=0, which='LM')
    else:
        eigenvalues, eigenvectors = eigh(K, M=M)

    if convert_to_hz:
        eigenvalues = np.sqrt(eigenvalues) / (2 * np.pi)

    return eigenvalues, eigenvectors


import numpy as np
import scipy.linalg
from scipy.sparse.linalg import eigsh
from scipy import sparse
import time

def matrices_k_e_timoshenko(l, E, I, A, nu=0.3, k_s=5/6):
    """
    Stiffness matrix for a beam element (Timoshenko theory).
    """
    G = E/(2*(1+nu))
    phi = (12*E*I)/(k_s*A*G*l**2)

    if type(l) == np.ndarray:
        ones = np.ones_like(l)
    else:
        ones = 1

    K = E*I/(l**3*(1+phi)) * np.array([[12*ones,  6*l,          -12*ones,  6*l],
                                    [6*l, (4+phi)*l**2, -6*l, (2-phi)*l**2],
                                    [-12*ones, -6*l,         12*ones,   -6*l],
                                    [6*l, (2-phi)*l**2, -6*l, (4+phi)*l**2]])

    return K

def matrices_m_e(l, m):
    """
    Mass matrix of a beam element.

    """
    if type(l) == np.ndarray:
        ones = np.ones_like(l)
        M = (m)/420 * np.array([[156*ones,   22*l,    54*ones,    -13*l],
                                [22*l,  4*l**2,  13*l,  -3*l**2],
                                [54*ones,    13*l,    156*ones,   -22*l],
                                [-13*l, -3*l**2, -22*l, 4*l**2]])
    else:
        M = (m)/420 * np.array([[156,   22*l,    54,    -13*l],
                                [22*l,  4*l**2,  13*l,  -3*l**2],
                                [54,    13*l,    156,   -22*l],
                                [-13*l, -3*l**2, -22*l, 4*l**2]])
    return M

def matrices_k_e(l, EI):
    """
    Stiffness matrix of a beam element.

    """
    if type(l) == np.ndarray:
        ones = np.ones_like(l)
        K = (EI)/l**3 * np.array([[12*ones,  6*l,    -12*ones,  6*l],
                                [6*l, 4*l**2, -6*l, 2*l**2],
                                [-12*ones, -6*l,   12*ones,   -6*l],
                                [6*l, 2*l**2, -6*l, 4*l**2]])
    
    else:
        K = (EI)/l**3 * np.array([[12,  6*l,    -12,  6*l],
                               [6*l, 4*l**2, -6*l, 2*l**2],
                               [-12, -6*l,   12,   -6*l],
                               [6*l, 2*l**2, -6*l, 4*l**2]])
    return K



class Beam:
    def __init__(self, org, conec, length, width, height, density, Young, n_nodes=None, added_masses=None, mass_locations=None,
                mode='EB'):
        """
        
        Parameters
        ----------
        org : array_like
            Organization matrix.
        conec : array_like
            Connectivity matrix.
        length : array_like
            Lengths of the beam elements.
        width : float
            Width of the beam.
        height : float
            Height of the beam.
        mass : array_like
            Masses of the beam elements.
        Young : array_like
            Young's modulus of each beam.
        n_nodes : int, optional
            Number of nodes to construct org and conec if not given.
        mode : str, optional
            The mode of assembly. Default is "EB".
             - "EB": Euler-Bernoulli beam theory.
             - "T": Timoshenko beam theory.
        """
        if org is None and n_nodes is not None:
            org = np.linspace(0, np.sum(length), n_nodes)
            conec = np.array([[i, i+1] for i in range(org.shape[0]-1)])
        
        self.org = org
        self.conec = conec

        self.n_dof_node = 2
        self.el_nodes = 2
        self.n_elements = n_nodes - 1
        self.n_dof = self.n_dof_node * n_nodes

        self.length = length
        self.width = width
        self.height = height
        self.mass = width * height * np.array(length) * np.array(density)
        self.Young = Young
        self.area = self.width*self.height

        self.I = (self.height**3*self.width)/12
        self.EI = np.array(self.Young) * self.I

        self.added_masses = added_masses
        self.mass_locations = mass_locations

        self.construct_loce()
        self.assemble(mode=mode)

        if self.added_masses is not None and self.mass_locations is not None:
            self.add_mass()

    def construct_loce(self):
        """
        Construct LOCE matrix from CONEC.
        """
        self.loce = []
        insert = np.arange(self.n_dof_node)
        conec1 = self.conec.flatten()
        for node in conec1:
            self.loce.append(node * self.n_dof_node + insert)

        self.loce = np.asarray(self.loce).flatten()
        self.loce = self.loce.reshape(
            self.conec.shape[0], self.n_dof_node * self.el_nodes)

    def assemble(self, mode="EB"):
        """Assemble mass and stiffness matrices.

        Parameters
        ----------
        mode : str, optional
        The mode of assembly. Default is "EB".
         - "EB": Euler-Bernoulli beam theory.
         - "T": Timoshenko beam theory.
        """
        # self.Ms = np.zeros((self.n_elements, self.n_dof, self.n_dof))
        # self.Ks = np.zeros((self.n_elements, self.n_dof, self.n_dof))

        self.M = np.zeros((self.n_dof, self.n_dof))
        self.K = np.zeros((self.n_dof, self.n_dof))

        self.Ms1 = matrices_m_e(self.length, self.mass)
        if mode == "EB":
            self.Ks1 = matrices_k_e(self.length, self.EI)
        elif mode == "Timoshenko":
            self.Ks1 = matrices_k_e_timoshenko(self.length, self.Young, self.I, self.area)

        for i in range(self.n_elements):
            # self.Ms[i][np.ix_(self.loce[i], self.loce[i])] = self.Ms1[:, :, i]
            # self.Ks[i][np.ix_(self.loce[i], self.loce[i])] = self.Ks1[:, :, i]

            self.M[np.ix_(self.loce[i], self.loce[i])] += self.Ms1[:, :, i]
            self.K[np.ix_(self.loce[i], self.loce[i])] += self.Ks1[:, :, i]

    def add_mass(self):
        for i, m in zip(self.mass_locations, self.added_masses):
            # self.M[np.ix_(self.loce[i], self.loce[i])][0, 0] += m
            self.M[self.loce[i][0], self.loce[i][0]] += m

    def solve(self, lanczos=True, n=10):
        """Solve eigen problem."""
        # if lanczos is True:
        # try:
        #     K = sparse.csc_matrix(self.K)
        #     M = sparse.csc_matrix(self.M)

        #     eigval, eigvec = eigsh(K, M=M, k=n, sigma=0, which='LM')
        #     # print('lanczos')
        # # else:
        # except:
        #     print('full')
        #     eigval, eigvec = scipy.linalg.eig(self.K, self.M)

        try:
            K = sparse.csc_matrix(self.K)
            M = sparse.csc_matrix(self.M)

            eigval, eigvec = eigsh(K, M=M, k=n, sigma=0, which='LM')
        except:
            K = sparse.csc_matrix(self.K)
            M = sparse.csc_matrix(self.M)

            eigval, eigvec = eigsh(K, M=M, k=n, sigma=100, which='LM')
            

        sort_inds = np.argsort(eigval) # sortiramo po velikosti
        eigval = eigval[sort_inds]
        eigvec = eigvec[:, sort_inds]

        eig_omega = np.sqrt(eigval)
        eig_omega[eig_omega != eig_omega] = 0  # nan element are 0
        self.nat_freq = np.real(eig_omega / (2*np.pi))
        self.A = eigvec

        return self.nat_freq, np.real(self.A)







import os
import sys
import pickle
import meshio
from pyvistaqt import BackgroundPlotter
import pyvista as pv
import numpy as np
from scipy.sparse.linalg import eigsh

# os.environ['QT_QPA_PLATFORM'] = 'xcb' # if you run wayland and pyqt6, uncomment this

current_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(current_path)

from sdypy.model import Shell


# Load the mesh
try:
    mesh = meshio.read("examples/data/L_bracket.msh")
except:
    mesh = meshio.read("data/L_bracket.msh")

# extract nodes and elements from mesh
nodes = mesh.points / 1000
elements = []
for cells in mesh.cells:
    if cells.type == "quad":
        elements.append(cells.data)
elements = np.vstack(elements)

#  Properties
thickness = 0.001
# E = 210e9
E = 2.069e11
nu = 0.3
rho = 7829

# Compute the stiffness and mass matrices
mass_lumping = False
force_nonsingularity = True

shell_obj = Shell(nodes, elements, E, nu, rho, thickness, verbose=1, mass_lumping=mass_lumping, force_nonsingularity=force_nonsingularity)
K = shell_obj.K
M = shell_obj.M

# Compute the first 20 modes
eigenvalues, eigenvectors = eigsh(K, M=M, k=20, sigma=0, which='LM')

# Plot the modes
n_plots = 6
rows = int(np.floor(np.sqrt(n_plots)))
cols = int(n_plots // rows)
shape = (rows, cols)

try:
    # Initialize a plotter with 6 subplots (2 rows, 3 columns)
    plotter = BackgroundPlotter(shape=shape)

    mode_start = 3
    for ii in range(int(rows*cols)):
        mode_number = mode_start + ii
        modal_shape = eigenvectors[:, mode_number]

        # Extract translational components (ux, uy, uz) from modal shape
        num_nodes = nodes.shape[0]
        translational_modal_shape = np.zeros_like(nodes)

        for i in range(num_nodes):
            translational_modal_shape[i, 0] = modal_shape[i * 6]     # ux
            translational_modal_shape[i, 1] = modal_shape[i * 6 + 1] # uy
            translational_modal_shape[i, 2] = modal_shape[i * 6 + 2] # uz

        # Scale the modal shape for better visualization
        # scale_factor = 1e-2  # Adjust this as needed
        scale_factor = np.linalg.norm(nodes)/np.linalg.norm(modal_shape)
        displaced_nodes = nodes + scale_factor * translational_modal_shape

        # Compute the norm of the displacements for coloring
        displacement_norm = np.linalg.norm(translational_modal_shape, axis=1)

        # Select the subplot to work on
        plotter.subplot(ii // cols, ii % cols)

        # Create a mesh for the current mode
        mesh = pv.PolyData(displaced_nodes)

        # Add cells (elements) to the mesh
        faces = []
        for element in elements:
            face = np.hstack([[4], element])  # Add number of points per cell (4 for quadrilateral)
            faces.append(face)
        faces = np.hstack(faces).astype(np.int64)
        mesh.faces = faces

        # Add displacement norm as point data for coloring
        displacement_norm = np.linalg.norm(translational_modal_shape, axis=1)  # Assuming displacements is a 3D array
        mesh["Displacement"] = displacement_norm

        # Add the mesh to the plotter
        plotter.add_mesh(mesh, show_edges=True, scalars="Displacement", cmap="viridis", edge_color='black')
        plotter.add_text(f"Mode {mode_number} ({np.sqrt(eigenvalues[mode_number])/(2*np.pi):.2f} Hz)", font_size=12)
        plotter.show_grid()

    # Show the plotter with all subplots
    plotter.show()
    input()
except Exception as e:
    print("To show the output on a plot, install a QT background.")
    print(e)

# %%




import numpy as np
from scipy.sparse import coo_array
from scipy.sparse import diags_array
from collections.abc import Iterable
from tqdm import tqdm

import warnings

def get_normal_vector(nodes_el):
    """Based on the nodal coordinates of the element, calculate the normal vector and the
    orthogonal vectors to the normal vector.
    
    The code is based on MATLAB code and the book "Finite element procedures"
    by Bathe.

    Parameters
    ----------
    nodes_el : np.ndarray
        array of the nodal coordinates of the element, shaped (4, 3).
    """
    Vn = np.zeros((4, 3))
    V1 = np.zeros((4, 3))
    V2 = np.zeros((4, 3))

    # calculate normal at each node
    R = np.array([-1, 1, 1, -1])
    S = np.array([-1, -1, 1, 1])

    for i in range(4):
        # derivative of the shape functions in the parent domain
        dN1dr = -0.25 * (1 - S[i])
        dN2dr = 0.25 * (1 - S[i])
        dN3dr = 0.25 * (1 + S[i])
        dN4dr = -0.25 * (1 + S[i])

        dN1ds = -0.25 * (1 - R[i])
        dN2ds = -0.25 * (1 + R[i])
        dN3ds = 0.25 * (1 + R[i])
        dN4ds = 0.25 * (1 - R[i])

        # construct first two columns of Jacobian 
        Jm = np.array([[dN1dr, dN2dr, dN3dr, dN4dr],
                    [dN1ds, dN2ds, dN3ds, dN4ds]]) @ nodes_el

        # normal vector 3 at the point
        Vn[i, :] = np.cross(Jm[0, :], Jm[1, :]) / np.linalg.norm(np.cross(Jm[0, :], Jm[1, :]))  # normalise the vector

        if np.abs(np.round(Vn[i, 1], 4)) == 1.:  # i.e. Vn is aligned with ey
            V1[i, :] = np.array([0, 0, 1.])  # see 5.110b in Bathe
            V2[i, :] = np.array([1., 0, 0])
        else:
            V1[i, :] = np.cross([0, 1, 0], Vn[i, :]) / np.linalg.norm(np.cross([0, 1, 0], Vn[i, :]))
            V2[i, :] = np.cross(Vn[i, :], V1[i, :]) / np.linalg.norm(np.cross(Vn[i, :], V1[i, :]))

    return Vn, V1, V2

def get_constitutive_tensor(E, nu, k=5/6):
    """Get the constitutive tensor of the given material properties.
    
    Parameters
    ----------
    E : float
        Young's modulus
    nu : float
        poisson ratio
    k : float
        shear correction factor
    """
    D0 = (E / (1 - nu**2)) * np.array([[1, nu, 0, 0, 0, 0],
                                    [nu, 1, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, (1 - nu) / 2, 0, 0],
                                    [0, 0, 0, 0, (k) * (1 - nu) / 2, 0],
                                    [0, 0, 0, 0, 0, (k) * (1 - nu) / 2]])
    
    return D0

def shape_functions(xi, eta):
    """Calculate the shape functions for the given xi and eta coordinates."""
    N = np.array([
        (1 - xi) * (1 - eta) / 4,
        (1 + xi) * (1 - eta) / 4,
        (1 + xi) * (1 + eta) / 4,
        (1 - xi) * (1 + eta) / 4
    ])
    return N

def shape_function_derivatives(xi, eta):
    """Calculate the derivatives of the shape functions for the given xi and eta coordinates."""
    dN_dxi = np.array([
        [-(1 - eta) / 4, -(1 - xi) / 4],
        [(1 - eta) / 4, -(1 + xi) / 4],
        [(1 + eta) / 4, (1 + xi) / 4],
        [-(1 + eta) / 4, (1 - xi) / 4]
    ])

    return dN_dxi

def interpolation_matrix(N1, N2, N3, N4, t, thickness, V1, V2):
    """Construct the interpolation matrix for the given shape functions, thickness
    and vectors V1 and V2 (computed by ``get_normal_vector``).
    
    Parameters
    ----------
    N1 : float
        shape function
    N2 : float
        shape function
    N3 : float
        shape function
    N4 : float
        shape function
    t : float
        coordinate in the direction of the thickness
    thickness : float
        thickness of the element
    V1 : array
        vector orthogonal to the normal vector
    V2 : array
        vector orthogonal to the normal vector
    """
    N = np.zeros((3, 20))
    N[:, :3] = np.diag([N1]*3)
    N[:, 5:8] = np.diag([N2]*3)
    N[:, 10:13] = np.diag([N3]*3)
    N[:, 15:18] = np.diag([N4]*3)

    N[:, 3] = -(t/2) * thickness * N1 * V2[0, :]
    N[:, 4] = (t/2) * thickness * N1 * V1[0, :]
    N[:, 8] = -(t/2) * thickness * N2 * V2[1, :]
    N[:, 9] = (t/2) * thickness * N2 * V1[1, :]
    N[:, 13] = -(t/2) * thickness * N3 * V2[2, :]
    N[:, 14] = (t/2) * thickness * N3 * V1[2, :]
    N[:, 18] = -(t/2) * thickness * N4 * V2[3, :]
    N[:, 19] = (t/2) * thickness * N4 * V1[3, :]
    return N

def construct_Brr(dN1dr, dN2dr, dN3dr, dN4dr, g1, g2, t):
    g1_1, g1_2, g1_3, g1_4 = g1
    g2_1, g2_2, g2_3, g2_4 = g2
    Brr = np.array([
        np.hstack((dN1dr * np.array([1, 0, 0, t * g1_1[0], t * g2_1[0]]),
                dN2dr * np.array([1, 0, 0, t * g1_2[0], t * g2_2[0]]),
                dN3dr * np.array([1, 0, 0, t * g1_3[0], t * g2_3[0]]),
                dN4dr * np.array([1, 0, 0, t * g1_4[0], t * g2_4[0]]))),
        np.hstack((dN1dr * np.array([0, 1, 0, t * g1_1[1], t * g2_1[1]]),
                dN2dr * np.array([0, 1, 0, t * g1_2[1], t * g2_2[1]]),
                dN3dr * np.array([0, 1, 0, t * g1_3[1], t * g2_3[1]]),
                dN4dr * np.array([0, 1, 0, t * g1_4[1], t * g2_4[1]]))),
        np.hstack((dN1dr * np.array([0, 0, 1, t * g1_1[2], t * g2_1[2]]),
                dN2dr * np.array([0, 0, 1, t * g1_2[2], t * g2_2[2]]),
                dN3dr * np.array([0, 0, 1, t * g1_3[2], t * g2_3[2]]),
                dN4dr * np.array([0, 0, 1, t * g1_4[2], t * g2_4[2]])))
    ])
    return Brr

def construct_Bss(dN1ds, dN2ds, dN3ds, dN4ds, g1, g2, t):
    g1_1, g1_2, g1_3, g1_4 = g1
    g2_1, g2_2, g2_3, g2_4 = g2
    Bss = np.array([
        np.hstack((dN1ds * np.array([1, 0, 0, t * g1_1[0], t * g2_1[0]]),
                dN2ds * np.array([1, 0, 0, t * g1_2[0], t * g2_2[0]]),
                dN3ds * np.array([1, 0, 0, t * g1_3[0], t * g2_3[0]]),
                dN4ds * np.array([1, 0, 0, t * g1_4[0], t * g2_4[0]]))),
        np.hstack((dN1ds * np.array([0, 1, 0, t * g1_1[1], t * g2_1[1]]),
                dN2ds * np.array([0, 1, 0, t * g1_2[1], t * g2_2[1]]),
                dN3ds * np.array([0, 1, 0, t * g1_3[1], t * g2_3[1]]),
                dN4ds * np.array([0, 1, 0, t * g1_4[1], t * g2_4[1]]))),
        np.hstack((dN1ds * np.array([0, 0, 1, t * g1_1[2], t * g2_1[2]]),
                dN2ds * np.array([0, 0, 1, t * g1_2[2], t * g2_2[2]]),
                dN3ds * np.array([0, 0, 1, t * g1_3[2], t * g2_3[2]]),
                dN4ds * np.array([0, 0, 1, t * g1_4[2], t * g2_4[2]])))
    ])
    return Bss

def construct_Btt(N1, N2, N3, N4, g1, g2, t):
    g1_1, g1_2, g1_3, g1_4 = g1
    g2_1, g2_2, g2_3, g2_4 = g2
    Btt = np.array([
        np.hstack((N1 * np.array([0, 0, 0, g1_1[0], g2_1[0]]),
                N2 * np.array([0, 0, 0, g1_2[0], g2_2[0]]),
                N3 * np.array([0, 0, 0, g1_3[0], g2_3[0]]),
                N4 * np.array([0, 0, 0, g1_4[0], g2_4[0]]))),
        np.hstack((N1 * np.array([0, 0, 0, g1_1[1], g2_1[1]]),
                N2 * np.array([0, 0, 0, g1_2[1], g2_2[1]]),
                N3 * np.array([0, 0, 0, g1_3[1], g2_3[1]]),
                N4 * np.array([0, 0, 0, g1_4[1], g2_4[1]]))),
        np.hstack((N1 * np.array([0, 0, 0, g1_1[2], g2_1[2]]),
                N2 * np.array([0, 0, 0, g1_2[2], g2_2[2]]),
                N3 * np.array([0, 0, 0, g1_3[2], g2_3[2]]),
                N4 * np.array([0, 0, 0, g1_4[2], g2_4[2]])))
    ])
    return Btt

def construct_rotation_tensor(gr, gs, gt):
    # Calculate the local coordinates
    e3 = gt / np.linalg.norm(gt)
    e1 = np.cross(gs, e3) / np.linalg.norm(np.cross(gs, e3))
    e2 = np.cross(e3, e1)

    # Calculate the contravariant basis
    gR = np.cross(gs, gt) / np.dot(gr, np.cross(gs, gt))
    gS = np.cross(gt, gr) / np.dot(gr, np.cross(gs, gt))
    gT = np.cross(gr, gs) / np.dot(gr, np.cross(gs, gt))

    # Generate the rotation tensor - 5.120 in Bathe
    l1, l2, l3 = np.dot(gR, e1), np.dot(gR, e2), np.dot(gR, e3)
    m1, m2, m3 = np.dot(gS, e1), np.dot(gS, e2), np.dot(gS, e3)
    n1, n2, n3 = np.dot(gT, e1), np.dot(gT, e2), np.dot(gT, e3)

    Q = np.array([
        [l1**2, m1**2, n1**2, l1*m1, m1*n1, n1*l1],
        [l2**2, m2**2, n2**2, l2*m2, m2*n2, n2*l2],
        [l3**2, m3**2, n3**2, l3*m3, m3*n3, n3*l3],
        [2*l1*l2, 2*m1*m2, 2*n1*n2, l1*m2 + l2*m1, m1*n2 + m2*n1, n1*l2 + n2*l1],
        [2*l2*l3, 2*m2*m3, 2*n2*n3, l2*m3 + l3*m2, m2*n3 + m3*n2, n2*l3 + n3*l2],
        [2*l3*l1, 2*m3*m1, 2*n3*n1, l3*m1 + l1*m3, m3*n1 + m1*n3, n3*l1 + n1*l3]
    ])
    return Q

def construct_rotational_matrix(V1, V2, Vn):
    # Rotate stiffness matrix
    T = np.zeros((24, 24))

    for n_index in range(4):
        transf = np.block([[np.eye(3), np.zeros((3, 3))],
                        [np.zeros((3, 3)), np.array([V1[n_index], V2[n_index], Vn[n_index]])]])
        T[n_index * 6: (n_index+1) * 6, n_index * 6:(n_index+1) * 6] = transf

    return T

def lump_mass_matrix(M):
    lumped_masses = M.sum(axis=1)  # Sum of rows
    lumped_mass_matrix = diags_array(lumped_masses, shape=M.shape)  # Create diagonal matrix
    return lumped_mass_matrix

# %%

def MITC4_element(E, nu, rho, thickness, nodes_el):
    """The computation of stiffness and mass matrices for each element.
    
    Using the MITC4 elements. The code is based on the MatLab code and on the
    Finite Element Procedures book by Bathe.
    """
    warnings.warn("MITC4_element is deprecated. Use Shell class instead.", DeprecationWarning)
    # Calculate normal at each node
    Vn, V1, V2 = get_normal_vector(nodes_el)
    
    # see eq 5.114 in Bathe
    g1 = -0.5 * thickness * V2
    g2 = 0.5 * thickness * V1

    # constitutive tensor in local coordinate system
    D0 = get_constitutive_tensor(E, nu)

    # perform full gauss integration (2 by 2 by 2 integration)
    W_gauss = np.array([1, 1])
    points_gauss = np.array([1/np.sqrt(3), -1/np.sqrt(3)])

    #  Compute K and M
    K_el = np.zeros((24, 24))
    M_el = np.zeros((24, 24))
    Volume_el = 0

    for r in points_gauss: # xi
        for s in points_gauss: # eta 
            for t in points_gauss: # zeta

                # interpolation functions
                N1, N2, N3, N4 = shape_functions(r, s)

                # derivative of the interpolation functions
                dN = shape_function_derivatives(r, s)
                dN1dr, dN1ds = dN[0]
                dN2dr, dN2ds = dN[1]
                dN3dr, dN3ds = dN[2]
                dN4dr, dN4ds = dN[3]
                
                # assemble the interpolation matrix (used to get the mass matrix)
                N = interpolation_matrix(N1, N2, N3, N4, t, thickness, V1, V2)

                # Constructing Brr
                Brr = construct_Brr(dN1dr, dN2dr, dN3dr, dN4dr, g1, g2, t)

                # Constructing Bss
                Bss = construct_Bss(dN1ds, dN2ds, dN3ds, dN4ds, g1, g2, t)

                # Constructing Btt
                Btt = construct_Btt(N1, N2, N3, N4, g1, g2, t)

                # Initialize GN with zeros
                GN = np.zeros((3, 4))

                # Fill GN with values
                GN[:2, :] = np.array([
                    [dN1dr, dN2dr, dN3dr, dN4dr],
                    [dN1ds, dN2ds, dN3ds, dN4ds]
                ])

                # Calculate GN_dash
                GN_dash = GN * 0.5 * t * thickness
                GN_dash[2, :] = np.array([N1, N2, N3, N4]) * 0.5 * thickness

                # Calculate the Jacobian
                J = np.dot(GN, nodes_el) + np.dot(GN_dash, Vn)
                detJ = np.linalg.det(J)

                # Covariant bases
                gr = J[0, :]
                gs = J[1, :]
                gt = J[2, :]

                # Calculate the direct strain components
                Err = np.dot(gr, Brr)
                Ess = np.dot(gs, Bss)
                Ett = np.dot(gt, Btt)  # This should be zero
                Ers = 0.5 * (np.dot(gs, Brr) + np.dot(gr, Bss))

                # Coordinates r1 and r2 for points A, B, C, D
                r1 = [0, -1, 0, 1]  # order A B C D
                r2 = [1, 0, -1, 0]

                # Initialize strain results
                Ert_A, Est_B, Ert_C, Est_D = None, None, None, None

                for i in range(4):
                    # Interpolation functions
                    N1, N2, N3, N4 = shape_functions(r1[i], r2[i])

                    # Derivative of the interpolation functions
                    dN = shape_function_derivatives(r1[i], r2[i])
                    dN1dr, dN1ds = dN[0]
                    dN2dr, dN2ds = dN[1]
                    dN3dr, dN3ds = dN[2]
                    dN4dr, dN4ds = dN[3]

                    # Constructing Brr
                    Brr = construct_Brr(dN1dr, dN2dr, dN3dr, dN4dr, g1, g2, t)

                    # Constructing Bss
                    Bss = construct_Bss(dN1ds, dN2ds, dN3ds, dN4ds, g1, g2, t)

                    # Constructing Btt
                    Btt = construct_Btt(N1, N2, N3, N4, g1, g2, t)

                    # Constructing GN_edge and GN_dash_edge
                    GN_edge = np.zeros((3, 4))
                    GN_edge[:2, :] = np.array([
                        [dN1dr, dN2dr, dN3dr, dN4dr],
                        [dN1ds, dN2ds, dN3ds, dN4ds]
                    ])

                    GN_dash_edge = GN_edge * 0.5 * t * thickness
                    GN_dash_edge[2, :] = np.array([N1, N2, N3, N4]) * 0.5 * thickness

                    # Jacobian
                    J_edge = np.dot(GN_edge, nodes_el) + np.dot(GN_dash_edge, Vn)

                    # Covariant bases
                    gr_edge = J_edge[0, :]
                    gs_edge = J_edge[1, :]
                    gt_edge = J_edge[2, :]

                    if i == 0:  # at A
                        Ert_A = 0.5 * (np.dot(gr_edge, Btt) + np.dot(gt_edge, Brr))
                    elif i == 1:  # at B
                        Est_B = 0.5 * (np.dot(gs_edge, Btt) + np.dot(gt_edge, Bss))
                    elif i == 2:  # at C
                        Ert_C = 0.5 * (np.dot(gr_edge, Btt) + np.dot(gt_edge, Brr))
                    else:  # at D
                        Est_D = 0.5 * (np.dot(gs_edge, Btt) + np.dot(gt_edge, Bss))

                # Perform the interpolation for the transverse strains
                Ert = 0.5 * (1 + s) * Ert_A + 0.5 * (1 - s) * Ert_C
                Est = 0.5 * (1 + r) * Est_D + 0.5 * (1 - r) * Est_B

                # Assemble the B vector
                B = np.vstack((Err, Ess, Ett, Ers * 2, Est * 2, Ert * 2))

                # # Generate the rotation tensor - 5.120 in Bathe
                Q = construct_rotation_tensor(gr, gs, gt)

                # Rotate the constitutive tensor
                D = np.dot(np.dot(Q.T, D0), Q)

                # Add contribution of the element matrices
                dofstest = np.setdiff1d(np.arange(24), np.arange(5, 24, 6))
                K_el[np.ix_(dofstest, dofstest)] += np.dot(np.dot(B.T, D), B) * detJ
                M_el[np.ix_(dofstest, dofstest)] += np.dot(np.dot(N.T, rho * np.eye(N.shape[0])), N) * detJ

                # Total volume
                Volume_el += detJ
    return K_el, M_el, {"Volume": Volume_el, "V1": V1, "V2": V2, "Vn": Vn}


def MITC4_global(nodes, elements, E, nu, rho, thickness, verbose=0, mass_lumping=False, 
                 force_nonsingularity=False, mass_threshold=None):
    """Construct the global stiffness matrix and mass matrices.
    
    Parameters
    ----------
    nodes : array
        The nodes of the mesh.
    elements : array
        The elements of the mesh.
    E : float or iterable
        The Young's modulus of the material. If float, all elements are the same.
        If iterable, must be the same length as elements.
    nu : float or iterable
        The Poisson's ratio of the material. If float, all elements are the same.
        If iterable, must be the same length as elements.
    rho : float or iterable
        The density of the material. If float, all elements are the same.
        If iterable, must be the same length as elements.
    thickness : float or iterable
        The thickness of the material. If float, all elements are the same.
        If iterable, must be the same length as elements.
    verbose : bool
        If True, show the progress bar.
    mass_lumping : bool
        If True, perform mass lumping.
    force_nonsingularity : bool
        If True, add a small value to the diagonal of the mass matrix to force nonsingularity.
    mass_threshold : float
        The threshold for the mass matrix. Elements of the matrix below this threshold are set to zero.
    """
    warnings.warn("MITC4_global is deprecated. Use Shell class instead.", DeprecationWarning)
    n_nodes = nodes.shape[0]
    n_el = elements.shape[0]

    dofs_per_node = 6
    nDOF = n_nodes * dofs_per_node

    # check if E is iterable
    if not isinstance(E, Iterable):
        E = np.array([E] * n_el)
    elif len(E) != n_el:
        raise ValueError("E must be the same length as elements.")

    if not isinstance(rho, Iterable):
        rho = np.array([rho] * n_el)
    elif len(rho) != n_el:
        raise ValueError("rho must be the same length as elements.")

    if not isinstance(thickness, Iterable):
        thickness = np.array([thickness] * n_el)
    elif len(thickness) != n_el:
        raise ValueError("nu must be the same length as elements.")
    
    if not isinstance(nu, Iterable):
        nu = np.array([nu] * n_el)
    elif len(nu) != n_el:
        raise ValueError("nu must be the same length as elements.")
    
    n_entries_in_matrix = (elements.shape[1] * dofs_per_node)**2 # 24 * 24 for 4 node, 6 dof per node element
    iIndex = np.zeros(n_el * n_entries_in_matrix, dtype=int)
    jIndex = np.zeros(n_el * n_entries_in_matrix, dtype=int)
    kAll = np.zeros(n_el * n_entries_in_matrix)
    mAll = np.zeros(n_el * n_entries_in_matrix)
    V_All = np.zeros(n_el)  # Initialize the volume vector

    if verbose:
        pbar = lambda x: tqdm(x)
        print("Constructing the global stiffness matrix and mass matrix...")
    else:
        pbar = lambda x: x

    for el, element in enumerate(pbar(elements)):
        nodes_el = nodes[element]
        
        # Get the element stiffness and mass matrix
        K_el, M_el, data_dict = MITC4_element(E[el], nu[el], rho[el], thickness[el], nodes_el)
        
        # Extract normal and orthogonal vectors
        V1 = data_dict["V1"]
        V2 = data_dict["V2"]
        Vn = data_dict["Vn"]

        # Store the volume
        V_All[el] = data_dict["Volume"]

        # --------------------------------------------------------
        # Add the drilling dofs -> rather use the method suggested in bathe book?
        Ketest = np.diag(K_el)
        
        minKetest = np.min(Ketest[Ketest != 0])
        K_el[5::6, 5::6] = (minKetest * 1e-5) * np.eye(4) 
        # I noticed that if the division is by 1000, the mode shape ordering is 
        # not ok (first deformation mode shape is before two of the rigid, rotational mode shapes)
        # --------------------------------------------------------

        # get the dofs for the element
        n1, n2, n3, n4 = element
        dofs = np.concatenate([np.arange(n1 * 6, (n1+1) * 6),
                            np.arange(n2 * 6, (n2+1) * 6),
                            np.arange(n3 * 6, (n3+1) * 6),
                            np.arange(n4 * 6, (n4+1) * 6)])
        
        # Get the rotation matrix for stiffness and mass matrix
        T = construct_rotational_matrix(V1, V2, Vn)

        # Rotate the stiffness and mass matrix
        K_el = np.dot(np.dot(T.T, K_el), T)
        M_el = np.dot(np.dot(T.T, M_el), T)

        # Perform mass lumping
        if mass_lumping:
            M_el = np.diag(np.sum(M_el, axis=1))

        # Add the element matrices to the global matrices
        nEntry = 24 * 24  # Number of entries in the stiffness matrix
        start_idx = el * nEntry
        end_idx = (el+1) * nEntry

        kAll[start_idx:end_idx] = K_el.flatten()
        mAll[start_idx:end_idx] = M_el.flatten()
        iIndex[start_idx:end_idx] = np.tile(dofs, 24)
        jIndex[start_idx:end_idx] = np.repeat(dofs, 24)


    # Create the sparse global stiffness and mass matrices
    # Create COO arrays (easier construction)
    K_coo = coo_array((kAll, (iIndex, jIndex)), shape=(nDOF, nDOF))
    M_coo = coo_array((mAll, (iIndex, jIndex)), shape=(nDOF, nDOF))

    if verbose:
        print("...matrices assembled.")

    # Convert COO arrays to CSR arrays (more efficient computation)
    K = K_coo.tocsr()
    M = M_coo.tocsr()

    if mass_threshold is not None:
        M[M < mass_threshold] = 0

    if force_nonsingularity:
        # add a small value to the diagonal of the mass matrix
        Mtest = np.abs(M.diagonal())
        minMtest = np.min(Mtest[Mtest != 0])
        # nonsingularity_factor = 1e-5
        nonsingularity_factor = 1e-4
        M = M + minMtest * nonsingularity_factor * diags_array(np.ones(M.shape[0]))

    K.eliminate_zeros()
    M.eliminate_zeros()

    return K, M

class Shell:
    def __init__(self, nodes, elements, E, nu, rho, thickness, verbose=0, 
                 mass_lumping=False, force_nonsingularity=False, mass_threshold=None):
        """Initialize the shell model.
        
        Parameters
        ----------
        nodes : array
            The nodes of the mesh.
        elements : array
            The elements of the mesh.
        E : float or iterable
            The Young's modulus of the material. If float, all elements are the same.
            If iterable, must be the same length as elements.
        nu : float or iterable
            The Poisson's ratio of the material. If float, all elements are the same.
            If iterable, must be the same length as elements.
        rho : float or iterable
            The density of the material. If float, all elements are the same.
            If iterable, must be the same length as elements.
        thickness : float or iterable
            The thickness of the material. If float, all elements are the same.
            If iterable, must be the same length as elements.
        verbose : bool
            If True, show the progress bar.
        mass_lumping : bool
            If True, perform mass lumping.
        force_nonsingularity : bool
            If True, add a small value to the diagonal of the mass matrix to force nonsingularity.
        mass_threshold : float
            The threshold for the mass matrix. Elements of the matrix below this threshold are set to zero.
        """
        self.E = E
        self.nu = nu
        self.rho = rho
        self.thickness = thickness
        self.nodes = nodes
        self.elements = elements
        self.verbose = verbose
        self.mass_lumping = mass_lumping
        self.force_nonsingularity = force_nonsingularity
        self.mass_threshold = mass_threshold

        self.n_nodes = self.nodes.shape[0]
        self.n_el = self.elements.shape[0]

        # check if self.E is iterable
        if not isinstance(self.E, Iterable):
            self.E = np.array([self.E] * self.n_el)
        elif len(self.E) != self.n_el:
            raise ValueError("self.E must be the same length as elements.")

        if not isinstance(self.rho, Iterable):
            self.rho = np.array([self.rho] * self.n_el)
        elif len(self.rho) != self.n_el:
            raise ValueError("rho must be the same length as elements.")

        if not isinstance(self.thickness, Iterable):
            self.thickness = np.array([self.thickness] * self.n_el)
        elif len(self.thickness) != self.n_el:
            raise ValueError("nu must be the same length as elements.")
        
        if not isinstance(self.nu, Iterable):
            self.nu = np.array([self.nu] * self.n_el)
        elif len(self.nu) != self.n_el:
            raise ValueError("nu must be the same length as elements.")
        
        # Construct the global stiffness and mass matrices
        self.construct_global_matrices()

    def construct_MITC4_matrices_for_element(self, nodes_el, el):
        """The computation of stiffness and mass matrices for each element.
        
        Using the MITC4 elements. The code is based on the MatLab code and on the
        Finite Element Procedures book by Bathe.
        """
        E = self.E[el]
        nu = self.nu[el]
        rho = self.rho[el]
        thickness = self.thickness[el]

        # Calculate normal at each node
        Vn, V1, V2 = get_normal_vector(nodes_el)
        
        # see eq 5.114 in Bathe
        g1 = -0.5 * thickness * V2
        g2 = 0.5 * thickness * V1

        # constitutive tensor in local coordinate system
        D0 = get_constitutive_tensor(E, nu)

        # perform full gauss integration (2 by 2 by 2 integration)
        W_gauss = np.array([1, 1])
        points_gauss = np.array([1/np.sqrt(3), -1/np.sqrt(3)])

        #  Compute K and M
        K_el = np.zeros((24, 24))
        M_el = np.zeros((24, 24))
        Volume_el = 0

        for r in points_gauss: # xi
            for s in points_gauss: # eta 
                for t in points_gauss: # zeta

                    # interpolation functions
                    N1, N2, N3, N4 = shape_functions(r, s)

                    # derivative of the interpolation functions
                    dN = shape_function_derivatives(r, s)
                    dN1dr, dN1ds = dN[0]
                    dN2dr, dN2ds = dN[1]
                    dN3dr, dN3ds = dN[2]
                    dN4dr, dN4ds = dN[3]
                    
                    # assemble the interpolation matrix (used to get the mass matrix)
                    N = interpolation_matrix(N1, N2, N3, N4, t, thickness, V1, V2)

                    # Constructing Brr
                    Brr = construct_Brr(dN1dr, dN2dr, dN3dr, dN4dr, g1, g2, t)

                    # Constructing Bss
                    Bss = construct_Bss(dN1ds, dN2ds, dN3ds, dN4ds, g1, g2, t)

                    # Constructing Btt
                    Btt = construct_Btt(N1, N2, N3, N4, g1, g2, t)

                    # Initialize GN with zeros
                    GN = np.zeros((3, 4))

                    # Fill GN with values
                    GN[:2, :] = np.array([
                        [dN1dr, dN2dr, dN3dr, dN4dr],
                        [dN1ds, dN2ds, dN3ds, dN4ds]
                    ])

                    # Calculate GN_dash
                    GN_dash = GN * 0.5 * t * thickness
                    GN_dash[2, :] = np.array([N1, N2, N3, N4]) * 0.5 * thickness

                    # Calculate the Jacobian
                    J = np.dot(GN, nodes_el) + np.dot(GN_dash, Vn)
                    detJ = np.linalg.det(J)

                    # Covariant bases
                    gr = J[0, :]
                    gs = J[1, :]
                    gt = J[2, :]

                    # Calculate the direct strain components
                    Err = np.dot(gr, Brr)
                    Ess = np.dot(gs, Bss)
                    Ett = np.dot(gt, Btt)  # This should be zero
                    Ers = 0.5 * (np.dot(gs, Brr) + np.dot(gr, Bss))

                    # Coordinates r1 and r2 for points A, B, C, D
                    r1 = [0, -1, 0, 1]  # order A B C D
                    r2 = [1, 0, -1, 0]

                    # Initialize strain results
                    Ert_A, Est_B, Ert_C, Est_D = None, None, None, None

                    for i in range(4):
                        # Interpolation functions
                        N1, N2, N3, N4 = shape_functions(r1[i], r2[i])

                        # Derivative of the interpolation functions
                        dN = shape_function_derivatives(r1[i], r2[i])
                        dN1dr, dN1ds = dN[0]
                        dN2dr, dN2ds = dN[1]
                        dN3dr, dN3ds = dN[2]
                        dN4dr, dN4ds = dN[3]

                        # Constructing Brr
                        Brr = construct_Brr(dN1dr, dN2dr, dN3dr, dN4dr, g1, g2, t)

                        # Constructing Bss
                        Bss = construct_Bss(dN1ds, dN2ds, dN3ds, dN4ds, g1, g2, t)

                        # Constructing Btt
                        Btt = construct_Btt(N1, N2, N3, N4, g1, g2, t)

                        # Constructing GN_edge and GN_dash_edge
                        GN_edge = np.zeros((3, 4))
                        GN_edge[:2, :] = np.array([
                            [dN1dr, dN2dr, dN3dr, dN4dr],
                            [dN1ds, dN2ds, dN3ds, dN4ds]
                        ])

                        GN_dash_edge = GN_edge * 0.5 * t * thickness
                        GN_dash_edge[2, :] = np.array([N1, N2, N3, N4]) * 0.5 * thickness

                        # Jacobian
                        J_edge = np.dot(GN_edge, nodes_el) + np.dot(GN_dash_edge, Vn)

                        # Covariant bases
                        gr_edge = J_edge[0, :]
                        gs_edge = J_edge[1, :]
                        gt_edge = J_edge[2, :]

                        if i == 0:  # at A
                            Ert_A = 0.5 * (np.dot(gr_edge, Btt) + np.dot(gt_edge, Brr))
                        elif i == 1:  # at B
                            Est_B = 0.5 * (np.dot(gs_edge, Btt) + np.dot(gt_edge, Bss))
                        elif i == 2:  # at C
                            Ert_C = 0.5 * (np.dot(gr_edge, Btt) + np.dot(gt_edge, Brr))
                        else:  # at D
                            Est_D = 0.5 * (np.dot(gs_edge, Btt) + np.dot(gt_edge, Bss))

                    # Perform the interpolation for the transverse strains
                    Ert = 0.5 * (1 + s) * Ert_A + 0.5 * (1 - s) * Ert_C
                    Est = 0.5 * (1 + r) * Est_D + 0.5 * (1 - r) * Est_B

                    # Assemble the B vector
                    B = np.vstack((Err, Ess, Ett, Ers * 2, Est * 2, Ert * 2))

                    # # Generate the rotation tensor - 5.120 in Bathe
                    Q = construct_rotation_tensor(gr, gs, gt)

                    # Rotate the constitutive tensor
                    D = np.dot(np.dot(Q.T, D0), Q)

                    # Add contribution of the element matrices
                    dofstest = np.setdiff1d(np.arange(24), np.arange(5, 24, 6))
                    K_el[np.ix_(dofstest, dofstest)] += np.dot(np.dot(B.T, D), B) * detJ
                    M_el[np.ix_(dofstest, dofstest)] += np.dot(np.dot(N.T, rho * np.eye(N.shape[0])), N) * detJ

                    # Total volume
                    Volume_el += detJ

        return K_el, M_el, {"Volume": Volume_el, "V1": V1, "V2": V2, "Vn": Vn}
    
    def construct_global_matrices(self):
        """Construct the global stiffness matrix and mass matrices."""
        dofs_per_node = 6
        nDOF = self.n_nodes * dofs_per_node
        
        n_entries_in_matrix = (self.elements.shape[1] * dofs_per_node)**2 # 24 * 24 for 4 node, 6 dof per node element
        iIndex = np.zeros(self.n_el * n_entries_in_matrix, dtype=int)
        jIndex = np.zeros(self.n_el * n_entries_in_matrix, dtype=int)
        kAll = np.zeros(self.n_el * n_entries_in_matrix)
        mAll = np.zeros(self.n_el * n_entries_in_matrix)
        V_All = np.zeros(self.n_el)  # Initialize the volume vector

        if self.verbose:
            def pbar(x):
                return tqdm(x)
            print("Constructing the global stiffness matrix and mass matrix...")
        else:
            def pbar(x):
                return x

        for el, element in enumerate(pbar(self.elements)):
            nodes_el = self.nodes[element]
            
            # Get the element stiffness and mass matrix
            # K_el, M_el, data_dict = MITC4_element(self.E[el], self.nu[el], self.rho[el], self.thickness[el], nodes_el)
            K_el, M_el, data_dict = self.construct_MITC4_matrices_for_element(nodes_el, el)
            
            # Extract normal and orthogonal vectors
            V1 = data_dict["V1"]
            V2 = data_dict["V2"]
            Vn = data_dict["Vn"]

            # Store the volume
            V_All[el] = data_dict["Volume"]

            # --------------------------------------------------------
            # Add the drilling dofs -> rather use the method suggested in bathe book?
            Ketest = np.diag(K_el)
            
            minKetest = np.min(Ketest[Ketest != 0])
            K_el[5::6, 5::6] = (minKetest * 1e-5) * np.eye(4) 
            # I noticed that if the division is by 1000, the mode shape ordering is 
            # not ok (first deformation mode shape is before two of the rigid, rotational mode shapes)
            # --------------------------------------------------------

            # get the dofs for the element
            n1, n2, n3, n4 = element
            dofs = np.concatenate([np.arange(n1 * 6, (n1+1) * 6),
                                np.arange(n2 * 6, (n2+1) * 6),
                                np.arange(n3 * 6, (n3+1) * 6),
                                np.arange(n4 * 6, (n4+1) * 6)])
            
            # Get the rotation matrix for stiffness and mass matrix
            T = construct_rotational_matrix(V1, V2, Vn)

            # Rotate the stiffness and mass matrix
            K_el = np.dot(np.dot(T.T, K_el), T)
            M_el = np.dot(np.dot(T.T, M_el), T)

            # Perform mass lumping
            if self.mass_lumping:
                M_el = np.diag(np.sum(M_el, axis=1))

            # Add the element matrices to the global matrices
            nEntry = 24 * 24  # Number of entries in the stiffness matrix
            start_idx = el * nEntry
            end_idx = (el+1) * nEntry

            kAll[start_idx:end_idx] = K_el.flatten()
            mAll[start_idx:end_idx] = M_el.flatten()
            iIndex[start_idx:end_idx] = np.tile(dofs, 24)
            jIndex[start_idx:end_idx] = np.repeat(dofs, 24)


        # Create the sparse global stiffness and mass matrices
        # Create COO arrays (easier construction)
        K_coo = coo_array((kAll, (iIndex, jIndex)), shape=(nDOF, nDOF))
        M_coo = coo_array((mAll, (iIndex, jIndex)), shape=(nDOF, nDOF))

        if self.verbose:
            print("...matrices assembled.")

        # Convert COO arrays to CSR arrays (more efficient computation)
        K = K_coo.tocsr()
        M = M_coo.tocsr()

        if self.mass_threshold is not None:
            M[M < self.mass_threshold] = 0

        if self.force_nonsingularity:
            # add a small value to the diagonal of the mass matrix
            Mtest = np.abs(M.diagonal())
            minMtest = np.min(Mtest[Mtest != 0])
            # nonsingularity_factor = 1e-5
            nonsingularity_factor = 1e-4
            M = M + minMtest * nonsingularity_factor * diags_array(np.ones(M.shape[0]))

        K.eliminate_zeros()
        M.eliminate_zeros()

        self.K = K
        self.M = M